PROJECT_FRAGMENT = '''
id
title
consensusTotCoverage
maxWorkerCount
minAgreement
minConsensusSize
interfaceCategory
inputType
consensusMark
honeypotMark
jsonInterface
numberOfRemainingAssets
numberOfAssets
numberOfAssetsWithSkippedLabels
numberOfReviewedAssets
numberOfLatestLabels
numberOfRoles
roles {
  id
  user { id, name, email }
  role
  consensusMark
  honeypotMark
  lastLabelingAt
  numberOfAnnotations
  numberOfLabels
  numberOfLabeledAssets
  totalDuration
}
'''



import time
from json import dumps, loads
from typing import List

from tqdm import tqdm

from ...helpers import GraphQLError, format_result
from ...queries.asset import QueriesAsset
from ...queries.project import QueriesProject
from .queries import (GQL_APPEND_TO_ROLES, GQL_CREATE_EMPTY_PROJECT,
                      GQL_DELETE_FROM_ROLES,
                      GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER,
                      GQL_UPDATE_PROJECT, GQL_UPDATE_PROPERTIES_IN_PROJECT,
                      GQL_UPDATE_PROPERTIES_IN_ROLE)


class MutationsProject:

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    def append_to_roles(self, project_id: str, user_email: str, role: str):
        """
        Add a user to a project

        Parameters
        ----------
        - project_id : str
        - user_email : str
            The email of the new user. This email is used as the unique identifier of the user.
        - role : str
            One of "ADMIN", "REVIEWER", "LABELER", "READER"

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'projectID': project_id,
            'userEmail': user_email,
            'role': role
        }
        result = self.auth.client.execute(GQL_APPEND_TO_ROLES, variables)
        return format_result('data', result)

    def update_properties_in_project(self, project_id: str,
                                     consensus_mark: float = None,
                                     consensus_tot_coverage: int = None,
                                     description: str = None,
                                     honeypot_mark: float = None,
                                     instructions: str = None,
                                     interface_category: str = 'IV2',
                                     input_type: str = None,
                                     json_interface: dict = None,
                                     min_consensus_size: int = None,
                                     number_of_assets: int = None,
                                     number_of_assets_with_empty_labels: int = None,
                                     number_of_latest_labels: int = None,
                                     number_of_remaining_assets: int = None,
                                     number_of_reviewed_assets: int = None,
                                     title: str = None):
        """
        Update properties of a project

        Parameters
        ----------
        - project_id : str
            Identifier of the project
        - consensus_mark : float, optional (default = None)
            Should be between 0 and 1
        - consensus_tot_coverage : int, optional (default = None)
            Should be between 0 and 100. It is the percentage of the dataset
            that will be annotated several times.
        - description : str, optional (default = None)
        - honeypot_mark : float, optional (default = None)
            Should be between 0 and 1
        - instructions : str, optional (default = None)
        - interface_category : str, optional (default = 'IV2')
            Always use 'IV2'
        - input_type : str, optional (default = None)
            Currently, one of {AUDIO, IMAGE, PDF, TEXT, URL, VIDEO, NA}
        - json_interface : dict, optional (default = None)
            The json parameters of the project, see Edit your interface.
        - min_consensus_size : int, optional (default = None)
            Number of people that will annotate the same asset, for consensus computation.
        - number_of_assets : int, optional (default = None)
            Defaults to 0
        - number_of_assets_with_empty_labels : int, optional (default = None)
            Defaults to 0
        - number_of_latest_labels : int, optional (default = None)
            Defaults to 0
        - number_of_remaining_assets : int, optional (default = None)
            Defaults to 0
        - number_of_reviewed_assets : int, optional (default = None)
            Defaults to 0
        - title : str, optional (default = None)

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'consensusMark': consensus_mark,
            'consensusTotCoverage': consensus_tot_coverage,
            'description': description,
            'honeypotMark': honeypot_mark,
            'instructions': instructions,
            'interfaceCategory': interface_category,
            'inputType': input_type,
            'jsonInterface': dumps(json_interface) if json_interface is not None else None,
            'minConsensusSize': min_consensus_size,
            'numberOfAssets': number_of_assets,
            'numberOfAssetsWithSkippedLabels': number_of_assets_with_empty_labels,
            'numberOfLatestLabels': number_of_latest_labels,
            'numberOfRemainingAssets': number_of_remaining_assets,
            'numberOfReviewedAssets': number_of_reviewed_assets,
            'projectID': project_id,
            'title': title
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_PROJECT, variables)
        return format_result('data', result)

    def create_empty_project(self, user_id: str):
        """
        Create an empty project

        Parameters
        ----------
        - user_id : str

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {'userID': user_id}
        result = self.auth.client.execute(GQL_CREATE_EMPTY_PROJECT, variables)
        return format_result('data', result)

    def update_project(self, project_id: str,
                       title: str,
                       description: str,
                       interface_category: str,
                       input_type: str = 'TEXT',
                       consensus_tot_coverage: int = 0,
                       min_consensus_size: int = 1,
                       max_worker_count: int = 4,
                       min_agreement: int = 66,
                       use_honey_pot: bool = False,
                       instructions: str = None):
        """
        Update a project

        Parameters
        ----------
        - project_id : str
        - title : str
        - description : str
        - interface_category : str
            Currently, one of
            {IV2, ALL, CLASSIFICATION, SENTIMENT, NER, NER_WITH_RELATIONS,
            RELATION_EXTRACTION, ENTITY_GROUPING, COREFERENCE_RESOLUTION,
            TRANSLATION, TRANSCRIPTION, IMAGE, MULTICLASS_IMAGE, CUSTOM,
            IMAGE_TO_GRAPH, IMAGE_WITH_SEARCH, IMAGE_TO_TEXT, MULTICLASS_TEXT_CLASSIFICATION,
            SINGLECLASS_TEXT_CLASSIFICATION, MULTICLASS_IMAGE_CLASSIFICATION,
            SINGLECLASS_IMAGE_CLASSIFICATION, VIDEO_CLASSIFICATION, NA}
        - input_type : str, optional (default = 'TEXT')
            Currently, one of {AUDIO, IMAGE, PDF, TEXT, URL, VIDEO, NA}
        - consensus_tot_coverage : int, optional (default = 0)
            Should be between 0 and 100. It is the percentage of the dataset
            that will be annotated several times.
        - min_consensus_size : int, optional (default = 1)
            Number of people that will annotate the same asset, for consensus computation.
        - max_worker_count : int, optional (default = 4)
            Maximum number of workers in the project
        - min_agreement : int, optional (default = 66)
            Should be a percentage (between 0 and 100)
        - use_honey_pot : bool, optional (default = False)
            Whether to compute honeypot in the project
        - instructions : str, optional (default = None)
            You can give instructions, they will be available to the annotators during the labeling process.
        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'projectID': project_id,
            'title': title,
            'description': description,
            'interfaceCategory': interface_category,
            'inputType': input_type,
            'consensusTotCoverage': consensus_tot_coverage,
            'minConsensusSize': min_consensus_size,
            'maxWorkerCount': max_worker_count,
            'minAgreement': min_agreement,
            'useHoneyPot': use_honey_pot,
            'instructions': instructions
        }
        result = self.auth.client.execute(GQL_UPDATE_PROJECT, variables)
        return format_result('data', result)

    def update_properties_in_role(self, role_id: str, project_id: str, user_id: str, role: str):
        """
        Update properties of a role

        You should be either an admin or a reviewer of the project, or an admin of the
        organization to be able to change the role of somebody.

        Parameters
        ----------
        - role_id : str
            Role identifier of the user. E.g. : 'to-be-deactivated'
        - project_id : str
            Identifier of the project
        - user_id : str
            The email or identifier of the user with updated role
        - role : str
            The new role. One of "ADMIN", "REVIEWER", "LABELER", "READER"

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'roleID': role_id,
            'projectID': project_id,
            'userID': user_id,
            'role': role
        }
        result = self.auth.client.execute(
            GQL_UPDATE_PROPERTIES_IN_ROLE, variables)
        return format_result('data', result)

    def delete_from_roles(self, role_id: str):
        """
        Delete users by their role_id

        Parameters
        ----------
        - role_id : str

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {'roleID': role_id}
        result = self.auth.client.execute(GQL_DELETE_FROM_ROLES, variables)
        return format_result('data', result)

    def update_properties_in_project_user(self, project_user_id: str,
                                          consensus_mark: float = None,
                                          honeypot_mark: float = None,
                                          json_interface: dict = None,
                                          number_of_labeled_assets: int = None,
                                          total_duration: int = None):
        """
        Update properties of a project-user tuple

        Parameters
        ----------
        - project_user_id : str
        - consensus_mark : float, optional (default = None)
            Should be between 0 and 1.
        - honeypot_mark : float, optional (default = None)
            Should be between 0 and 1.
        - json_interface : dict, optional (default = None)
            The json parameters of the project, see Edit your interface.
        - number_of_labeled_assets : int, optional (default = None)
            Number of assets the user labeled in the project.
        - total_duration : int, optional (default = None)
            Total time the user spent in the project.

        Returns
        -------
        - a result object which indicates if the mutation was successful, or an error message else.
        """
        variables = {
            'consensusMark': consensus_mark,
            'honeypotMark': honeypot_mark,
            'jsonInterface': dumps(json_interface),
            'numberOfLabeledAssets': number_of_labeled_assets,
            'projectUserID': project_user_id,
            'totalDuration': total_duration,
        }
        result = self.auth.client.execute(
            GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER, variables)
        return format_result('data', result)

    def force_project_kpis(self, project_id: str):
        """
        Compute KPIs for a project

        Parameters
        ----------
        - project_id : str

        Returns
        -------
        - None
        """
        _ = QueriesAsset(self.auth).assets(project_id=project_id)
        _ = QueriesProject(self.auth).projects(project_id=project_id)



from ...helpers import deprecate, format_result
from .queries import GQL_PROJECTS
from ...constants import NO_ACCESS_RIGHT


class QueriesProject:

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    def projects(self, project_id: str = None, search_query: str = None, skip: int = 0, first: int = 100):
        """
        Get projects with a search_query

        Parameters
        ----------
        search_query : str, optional (default = None)
            Returned projects have a title or a description that matches this string.
        skip : int, optional (default = 0)
            Number of projects to skip (they are ordered by their creation)
        first : int , optional (default = 100)
            Maximum number of projects to return

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.
        """
        variables = {
            'where': {
                'id': project_id,
                'searchQuery': search_query
            },
            'skip': skip,
            'first': first
        }
        result = self.auth.client.execute(GQL_PROJECTS, variables)
        return format_result('data', result)

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_projects used to fetch projects. It is now achievable with projects. It will be removed on June 1st.
        To fetch projects, use:
            > playground.projects(search_query=search_query, skip=skip, first=first)
        """)
    def get_projects(self, user_id: str, search_query: str = None, skip: int = 0, first: int = 100):
        return self.projects(search_query=search_query, skip=skip, first=first)

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_project used to fetch a project. It is now achievable with projects. It will be removed on June 1st.
        To fetch projects, use:
            > playground.projects(project_id=project_id)
        """)
    def get_project(self, project_id: str):
        projects = self.projects(project_id=project_id)
        assert len(projects) == 1, NO_ACCESS_RIGHT
        return projects[0]



from ...helpers import format_result
from .queries import GQL_PROJECT_USERS, GQL_PROJECT_USERS_WITH_KPIS


class QueriesProjectUser:

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    def project_users(self, email=None, id=None, organization_id=None, project_id=None, first=100, skip=0, with_kpis=False):
        """
        Return projects and their users (possibly with their KPIs)

        Parameters
        ----------
        - email : str, optional (default = None)
        - organization_id : str, optional (default = None)
        - project_id : str, optional (default = None)
        - first : int, optional (default = 100)
            Maximum number of users to return
        - skip : int, optional (default = 0)
            Number of project users to skip
        - with_kpis : bool, optional (default = False)
            Whether or not to compute kpis

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.
        """
        variables = {
            'first': first,
            'skip': skip,
            'where': {
                'id': id,
                'project': {
                    'id': project_id,
                },
                'user': {
                    'email': email,
                    'organization': {
                        'id': organization_id,
                    }
                },
            }
        }
        query = GQL_PROJECT_USERS_WITH_KPIS if with_kpis else GQL_PROJECT_USERS
        result = self.auth.client.execute(query, variables)
        return format_result('data', result)



import kili.mutations.asset
import kili.mutations.label
import kili.mutations.organization
import kili.mutations.project
import kili.mutations.user
import kili.queries.asset
import kili.queries.label
import kili.queries.organization
import kili.queries.project
import kili.queries.project_user
import kili.queries.user
import kili.subscriptions.label
from kili.constants import NO_ACCESS_RIGHT

from kili.helpers import deprecate


class Playground(
        kili.mutations.asset.MutationsAsset,
        kili.mutations.label.MutationsLabel,
        kili.mutations.organization.MutationsOrganization,
        kili.mutations.project.MutationsProject,
        kili.mutations.user.MutationsUser,
        kili.queries.asset.QueriesAsset,
        kili.queries.label.QueriesLabel,
        kili.queries.organization.QueriesOrganization,
        kili.queries.project.QueriesProject,
        kili.queries.project_user.QueriesProjectUser,
        kili.queries.user.QueriesUser,
        kili.subscriptions.label.SubscriptionsLabel):

    def __init__(self, auth=None):
        """Create an instance of KiliPlayground."""
        self.auth = auth
        super().__init__(auth)

    # Mutations Tool

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        Tools used to describe an interface. They are now called jsonInterface.
        To update jsonInterface, use:
            > playground.update_properties_in_project(project_id=project_id, json_interface=json_interface)
        """)
    def update_tool(self, project_id, json_settings):
        return self.update_properties_in_project(project_id, json_interface=json_settings)

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        Tools used to describe an interface. They are now called jsonInterface.
        To update jsonInterface, use:
            > playground.update_properties_in_project(project_id=project_id, json_interface=json_interface)
        """)
    def append_to_tools(self, project_id, json_settings):
        return self.update_properties_in_project(project_id, json_interface=json_settings)

    # Queries Tool

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        Tools used to describe an interface. They are now called jsonInterface.
        To query a jsonInterface, use:
            > playground.projects(project_id=project_id)
        """)
    def get_tools(self, project_id):
        projects = self.projects(project_id)
        assert len(projects) == 1, NO_ACCESS_RIGHT
        tool = {
            'jsonSettings': projects[0]['jsonInterface']
        }
        return [tool]


if __name__ == '__main__':
    """ Example of usage """
    from kili.authentication import KiliAuth
    from kili.playground import Playground
    kauth = KiliAuth()
    playground = Playground(kauth)
    assets = playground.assets(project_id="first-project")
    print(assets)



from .fragments import USER_FRAGMENT

GQL_USERS = f'''
query($where: UserWhere!, $first: PageSize!, $skip: Int!) {{
  data: users(where: $where, first: $first, skip: $skip) {{
    {USER_FRAGMENT}
  }}
}}
'''



import functools
import warnings
import base64
import re
from json import dumps, loads


class GraphQLError(Exception):
    def __init__(self, mutation, error):
        super().__init__(f'Mutation "{mutation}" failed with error: "{error}"')


def format_result(name, result):
    if 'errors' in result:
        raise GraphQLError(name, result['errors'])
    return format_json(result['data'][name])


def content_escape(content):
    return content.replace('\\', '\\\\').replace('\n', '\\n').replace('"', '\\"')


def get_data_type(path):
    if re.match(r'(jpg|jpeg)$', path.lower()):
        return 'image/png'


def encode_image(path):
    data_type = get_data_type(path)
    with open(path, 'rb') as image_file:
        return f'data:{data_type};base64,' + \
            base64.b64encode(image_file.read()).decode('ascii')


def is_url(path):
    return re.match(r'^(http://|https://)', path.lower())


def format_json(result):
    if result is None:
        return result
    if isinstance(result, list):
        return [format_json(elem) for elem in result]
    if isinstance(result, dict):
        for key, value in result.items():
            if key in ['jsonInterface', 'jsonMetadata', 'jsonResponse']:
                if value == '' or value is None:
                    result[key] = dict()
                else:
                    result[key] = loads(value)
            else:
                result[key] = format_json(value)
        return result
    return result


def deprecate(msg, type=DeprecationWarning):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(msg, type, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator



import os
import warnings

import requests

from . import __version__
from .graphql_client import GraphQLClient
from .mutations.user import signin

MAX_RETRIES = 20


class KiliAuth(object):
    """
    from kili.authentication import KiliAuth
    from kili.playground import Playground
    kauth = KiliAuth(email=email, password=password)
    playground = Playground(kauth)
    assets = playground.assets(project_id=project_id)
    """

    def __init__(self,
                 email=os.getenv('KILI_USER_EMAIL'),
                 password=os.getenv('KILI_USER_PASSWORD'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/graphql'):
        self.session = requests.Session()

        self.check_versions_match(api_endpoint)

        adapter = requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES)
        self.session.mount('https://', adapter)
        self.session.mount('http://', adapter)
        self.client = GraphQLClient(api_endpoint, self.session)
        auth_payload = signin(self.client, email, password)
        api_token = auth_payload['token']
        self.client.inject_token('Bearer: ' + api_token)
        self.user_id = auth_payload['user']['id']

    def __del__(self):
        self.session.close()

    @staticmethod
    def check_versions_match(api_endpoint):
        url = api_endpoint.replace('/graphql', '/version')
        response = requests.get(url).json()
        version = response['version']
        if version != __version__:
            message = 'Kili Playground version should match with Kili API version.\n' + \
                      f'Please install version: "pip install kili=={version}"'
            warnings.warn(message, UserWarning)



import time
from json import dumps
from typing import List

import pandas as pd
from tqdm import tqdm

from ...helpers import deprecate, format_result
from .queries import (GQL_ASSETS,
                      GQL_ASSETS_COUNT,
                      GQL_GET_NEXT_ASSET_FROM_LABEL,
                      GQL_GET_NEXT_ASSET_FROM_PROJECT)
from .fragments import ASSET_FRAGMENT_SIMPLIFIED
from ...constants import NO_ACCESS_RIGHT


class QueriesAsset:

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_asset used to fetch an asset from its ID. It is now achievable with assets.
        To fetch an asset from its ID, use:
            > playground.assets(asset_id=asset_id)
        """)
    def get_asset(self, asset_id: str):
        assets = self.assets(asset_id=asset_id)
        assert len(assets) == 1, NO_ACCESS_RIGHT
        return assets[0]

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_assets used to fetch assets. It is now achievable with assets.
        To fetch assets, use:
            > playground.assets(project_id=project_id)
        """)
    def get_assets(self, project_id: str):
        return self.assets(project_id=project_id)

    def assets(self, asset_id: str = None, project_id: str = None,
               skip: int = 0, first: int = None,
               external_id_contains: List[str] = None,
               status_in: List[str] = None,
               consensus_mark_gt: float = None,
               consensus_mark_lt: float = None,
               honeypot_mark_gt: float = None,
               honeypot_mark_lt: float = None,
               label_type_in: List[str] = None,
               label_author_in: List[str] = None,
               label_consensus_mark_gt: float = None,
               label_consensus_mark_lt: float = None,
               label_honeypot_mark_gt: float = None,
               label_honeypot_mark_lt: float = None,
               label_created_at_gt: float = None,
               label_created_at_lt: float = None,
               label_skipped: bool = None,
               format: str = None, disable_tqdm: bool = False,
               fragment=ASSET_FRAGMENT_SIMPLIFIED):
        """
        Get an array of assets from a project

        Parameters
        ----------
        - project_id : str
            Identifier of the project.
        - skip : int, optional (default = None)
            Number of assets to skip (they are ordered by their date of creation, first to last).
        - first : int, optional (default = None)
            Maximum number of assets to return.
        - external_id_contains : list of str, optional (default = None)
            Returned assets should have an external id that belongs to that list, if given.
        - status_in : list of str, optional (default = None)
            Returned assets should have a status that belongs to that list, if given.
            Possible choices : {'TODO', 'ONGOING', 'LABELED', 'REVIEWED'}
        - consensus_mark_gt : float, optional (default = None)
            Minimum amout of consensus for the asset.
        - consensus_mark_lt : float, optional (default = None)
            Maximum amout of consensus for the asset.
        - honeypot_mark_gt : float, optional (default = None)
            Minimum amout of honeypot for the asset.
        - honeypot_mark_lt : float, optional (default = None)
            Maximum amout of honeypot for the asset.
        - label_type_in : list of str, optional (default = None)
            Returned assets should have a label whose type belongs to that list, if given.
        - label_author_in : list of str, optional (default = None)
            Returned assets should have a label whose status belongs to that list, if given.
        - label_consensus_mark_gt, optional (default = None)
            Returned assets should have a label whose consensus is greater than this number.
        - label_consensus_mark_lt : float, optional (default = None)
            Returned assets should have a label whose consensus is lower than this number.
        - label_honeypot_mark_gt : float, optional (default = None)
            Returned assets should have a label whose honeypot is greater than this number.
        - label_honeypot_mark_lt : float, optional (default = None)
            Returned assets should have a label whose honeypot is lower than this number.
        - label_created_at_gt : float, optional (default = None)
            Returned assets should have a label whose creation date is greater than this date.
        - label_created_at_lt : float, optional (default = None)
            Returned assets should have a label whose creation date is lower than this date.
        - label_skipped : bool, optional (default = None)
            Returned assets should have a label which is skipped
        - format : str, optional (default = None)
            If equal to 'pandas', returns a pandas DataFrame
        - disable_tqdm : bool, optional (default = False)
        - fragment : str, optional (default = ASSET_FRAGMENT_SIMPLIFIED)


        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.
        """
        saved_args = locals()
        count_args = {k: v for (k, v) in saved_args.items()
                      if k not in ['skip', 'first', 'disable_tqdm', 'format', 'fragment', 'self']}
        number_of_assets_with_search = self.count_assets(**count_args)
        total = min(number_of_assets_with_search,
                    first) if first is not None else number_of_assets_with_search
        formatted_first = first if first else 100
        if total == 0:
            return []
        with tqdm(total=total, disable=disable_tqdm) as pbar:
            paged_assets = []
            while True:
                variables = {
                    'where': {
                        'id': asset_id,
                        'project': {
                            'id': project_id,
                        },
                        'externalIdIn': external_id_contains,
                        'statusIn': status_in,
                        'consensusMarkGte': consensus_mark_gt,
                        'consensusMarkLte': consensus_mark_lt,
                        'honeypotMarkGte': honeypot_mark_gt,
                        'honeypotMarkLte': honeypot_mark_lt,
                        'label': {
                            'typeIn': label_type_in,
                            'authorIn': label_author_in,
                            'consensusMarkGte': label_consensus_mark_gt,
                            'consensusMarkLte': label_consensus_mark_lt,
                            'honeypotMarkGte': label_honeypot_mark_gt,
                            'honeypotMarkLte': label_honeypot_mark_lt,
                            'createdAtGte': label_created_at_gt,
                            'createdAtLte': label_created_at_lt,
                            'skipped': label_skipped,
                        }
                    },
                    'skip': skip,
                    'first': formatted_first,
                }
                result = self.auth.client.execute(
                    GQL_ASSETS(fragment), variables)
                assets = format_result('data', result)
                if assets is None or len(assets) == 0 or (first is not None and len(paged_assets) == first):
                    if format == 'pandas':
                        return pd.DataFrame(paged_assets)
                    return paged_assets
                if first is not None:
                    assets = assets[:max(0, first - len(paged_assets))]
                paged_assets += assets
                skip += formatted_first
                pbar.update(len(assets))

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_assets_by_external_id used to fetch assets from an external_id. It is now achievable with assets.
        To fetch assets from and external_id, use:
            > playground.assets(project_id=project_id, external_id_contains=[external_id])
        """)
    def get_assets_by_external_id(self, project_id: str, external_id: str):
        return self.assets(project_id=project_id, external_id_contains=[external_id])

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        It will be removed on June 1st.
        """)
    def get_next_asset_from_label(self, label_asset_ids: List[str]):
        """
        Get next asset to label from previously labeled asset identifiers

        Parameters
        ----------
        - label_asset_ids : list of str
            The identifiers of the assets that have been labeled

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.
        """
        variables = {'labelAssetIDs': label_asset_ids}
        result = self.auth.client.execute(
            GQL_GET_NEXT_ASSET_FROM_LABEL, variables)
        return format_result('data', result)

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        It will be removed on June 1st.
        """)
    def get_next_asset_from_project(self, project_id: str):
        """
        Get next asset to label from project id

        Returns the next asset to label to the current user for the current project.

        Parameters
        ----------
        - project_id : str

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.
        """
        variables = {'projectID': project_id}
        result = self.auth.client.execute(
            GQL_GET_NEXT_ASSET_FROM_PROJECT, variables)
        return format_result('data', result)

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        export_assets used to fetch assets from a project. It is now achievable with assets.
        To fetch assets from a project, use:
            > playground.assets(project_id=project_id)
        """)
    def export_assets(self, project_id: str):
        return self.assets(project_id=project_id)

    def count_assets(self, asset_id: str = None,
                     project_id: str = None,
                     external_id_contains: List[str] = None,
                     status_in: List[str] = None,
                     consensus_mark_gt: float = None,
                     consensus_mark_lt: float = None,
                     honeypot_mark_gt: float = None,
                     honeypot_mark_lt: float = None,
                     label_type_in: List[str] = None,
                     label_author_in: List[str] = None,
                     label_consensus_mark_gt: float = None,
                     label_consensus_mark_lt: float = None,
                     label_honeypot_mark_gt: float = None,
                     label_honeypot_mark_lt: float = None,
                     label_created_at_gt: float = None,
                     label_created_at_lt: float = None,
                     label_skipped: bool = None):
        """
        Count and return the number of assets with the given parameters

        Parameters beginning with 'label_' apply to labels, others apply to assets.

        Parameters
        ----------
        - asset_id : str
            Identifier of the asset
        - project_id : str
            Identifier of the project
        - external_id_contains : list of str, optional (default = None)
            Returned assets should have an external id that belongs to that list, if given.
        - status_in : list of str, optional (default = None)
            Returned assets should have a status that belongs to that list, if given.
            Possible choices : {'TODO', 'ONGOING', 'LABELED', 'REVIEWED'}
        - consensus_mark_gt : float, optional (default = None)
            Minimum amout of consensus for the asset.
        - consensus_mark_lt : float, optional (default = None)
            Maximum amout of consensus for the asset.
        - honeypot_mark_gt : float, optional (default = None)
            Minimum amout of honeypot for the asset.
        - honeypot_mark_lt : float, optional (default = None)
            Maximum amout of consensus for the asset.
        - label_type_in : list of str, optional (default = None)
            Returned assets should have a label whose type belongs to that list, if given.
        - label_author_in : list of str, optional (default = None)
            Returned assets should have a label whose status belongs to that list, if given.
        - label_consensus_mark_gt : float, optional (default = None)
            Returned assets should have a label whose consensus is greater than this number.
        - label_consensus_mark_lt : float, optional (default = None)
            Returned assets should have a label whose consensus is lower than this number.
        - label_honeypot_mark_gt : float, optional (default = None)
            Returned assets should have a label whose honeypot is greater than this number.
        - label_honeypot_mark_lt : float, optional (default = None)
            Returned assets should have a label whose honeypot is lower than this number.
        - label_created_at_gt : float, optional (default = None)
            Returned assets should have a label whose creation date is greater than this date.
        - label_created_at_lt : float, optional (default = None)
            Returned assets should have a label whose creation date is lower than this date.
        - label_skipped : bool, optional (default = None)
            Returned assets should have a label which is skipped

        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.
        """
        variables = {
            'where': {
                'id': asset_id,
                'project': {
                    'id': project_id,
                },
                'externalIdIn': external_id_contains,
                'statusIn': status_in,
                'consensusMarkGte': consensus_mark_gt,
                'consensusMarkLte': consensus_mark_lt,
                'honeypotMarkGte': honeypot_mark_gt,
                'honeypotMarkLte': honeypot_mark_lt,
                'label': {
                    'typeIn': label_type_in,
                    'authorIn': label_author_in,
                    'consensusMarkGte': label_consensus_mark_gt,
                    'consensusMarkLte': label_consensus_mark_lt,
                    'honeypotMarkGte': label_honeypot_mark_gt,
                    'honeypotMarkLte': label_honeypot_mark_lt,
                    'createdAtGte': label_created_at_gt,
                    'createdAtLte': label_created_at_lt,
                    'skipped': label_skipped,
                }
            }
        }
        result = self.auth.client.execute(GQL_ASSETS_COUNT, variables)
        count = format_result('data', result)
        return count



from .fragments import ORGANIZATION_FRAGMENT

GQL_ORGANIZATIONS = f'''
query($where: OrganizationWhere!, $first: PageSize!, $skip: Int!) {{
  data: organizations(where: $where, first: $first, skip: $skip) {{
    {ORGANIZATION_FRAGMENT}
  }}
}}
'''



from .fragments import LABEL_FRAGMENT

GQL_LABELS = f'''
query ($where: LabelWhere!, $first: PageSize!, $skip: Int!) {{
  data: labels(where: $where, first: $first, skip: $skip) {{
    {LABEL_FRAGMENT}
  }}
}}
'''



from .fragments import ASSET_FRAGMENT

GQL_GET_ASSET = f'''
query($assetID: ID!) {{
  data: getAsset(assetID: $assetID) {{
    {ASSET_FRAGMENT}
  }}
}}
'''


def GQL_ASSETS(fragment):
    return f'''
query($where: AssetWhere!, $first: PageSize!, $skip: Int!) {{
  data: assets(where: $where, skip: $skip, first: $first) {{
    {fragment}
  }}
}}
'''


GQL_GET_NEXT_ASSET_FROM_LABEL = f'''
query($labelAssetIDs: [ID!]) {{
  data: getNextAssetFromLabel(labelAssetIDs: $labelAssetIDs) {{
    {ASSET_FRAGMENT}
  }}
}}
'''

GQL_GET_NEXT_ASSET_FROM_PROJECT = f'''
query($projectID: ID!) {{
  data: getNextAssetFromProject(projectID: $projectID) {{
    {ASSET_FRAGMENT}
  }}
}}
'''

GQL_ASSETS_COUNT = f'''
query($where: AssetWhere!) {{
  data: countAssets(where: $where)
}}
'''



PROJECT_USER_FRAGMENT = '''
id
activated
role
user {
  id
  email
}
'''

PROJECT_USER_FRAGMENT_WITH_KPIS = '''
id
activated
consensusMark
honeypotMark
lastLabelingAt
numberOfAnnotations
numberOfLabeledAssets
numberOfLabels
role
starred
totalDuration
user {
  id
  email
}
'''



from .fragments import PROJECT_USER_FRAGMENT, PROJECT_USER_FRAGMENT_WITH_KPIS

GQL_PROJECT_USERS = f'''
query($where: ProjectUserWhere!, $first: PageSize!, $skip: Int!) {{
  data: projectUsers(where: $where, first: $first, skip: $skip) {{
    {PROJECT_USER_FRAGMENT}
  }}
}}
'''

GQL_PROJECT_USERS_WITH_KPIS = f'''
query($where: ProjectUserWhere!, $first: PageSize!, $skip: Int!) {{
  data: projectUsers(where: $where, first: $first, skip: $skip) {{
    {PROJECT_USER_FRAGMENT_WITH_KPIS}
  }}
}}
'''



LOCK_FRAGMENT = '''
id
author {
  email
}
createdAt
lockType
'''

ASSET_FRAGMENT_AUTHOR = '''
id
email
'''

ASSET_FRAGMENT = f'''
id
externalId
content
isHoneypot
duration
consensusMark
honeypotMark
status
isUsedForConsensus
jsonMetadata
priority
labels {{
  id
  createdAt
  labelType
  jsonResponse
  isLatestLabelForUser
  numberOfAnnotations
  secondsToLabel
  totalSecondsToLabel
  honeypotMark
  skipped
  author {{
    {ASSET_FRAGMENT_AUTHOR}
  }}
}}
locks {{
  {LOCK_FRAGMENT}
}}
'''


ASSET_FRAGMENT_SIMPLIFIED = f'''
id
content
externalId
jsonMetadata
createdAt
updatedAt
isHoneypot
status
labels {{
  id
  author {{
    {ASSET_FRAGMENT_AUTHOR}
  }}
  labelType
  jsonResponse
  createdAt
  secondsToLabel
  totalSecondsToLabel
  honeypotMark
  isLatestLabelForUser
}}
'''



from .fragments import PROJECT_FRAGMENT

GQL_PROJECTS = f'''
query($where: ProjectWhere!, $first: PageSize!, $skip: Int!) {{
  data: projects(where: $where, first: $first, skip: $skip) {{
    {PROJECT_FRAGMENT}
  }}
}}
'''



import pandas as pd
from typing import List

from ...helpers import deprecate, format_result
from ..asset import QueriesAsset
from ..project import QueriesProject
from .queries import GQL_LABELS
from ...constants import NO_ACCESS_RIGHT


class QueriesLabel:

    def __init__(self, auth):
        """
        Initializes the subclass

        Parameters
        ----------
        - auth : KiliAuth object
        """
        self.auth = auth

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_label used to fetch labels from an asset_id and a user_id. It is now achievable with labels.
        To fetch labels from an asset_id and a user_id, use:
            > playground.labels(asset_id=asset_id, user_id=user_id)
        """)
    def get_label(self, asset_id: str, user_id: str):
        labels = self.labels(asset_id=asset_id, user_id=user_id)
        assert len(labels) == 1, NO_ACCESS_RIGHT
        return labels[0]

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_latest_labels_for_user used to fetch labels from a project_id and a user_id. It is now achievable with labels.
        To fetch labels from a project_id and a user_id, use:
            > playground.labels(project_id=project_id, user_id=user_id)
        """)
    def get_latest_labels_for_user(self, project_id: str, user_id: str):
        return self.labels(project_id=project_id, user_id=user_id)

    @deprecate(
        """
        This method is deprecated since: 30/04/2020.
        This method will be removed after: 30/05/2020.
        get_latest_labels used to fetch labels from a project_id. It is now achievable with labels.
        To fetch labels from a project_id, use:
            > playground.labels(project_id=project_id, first=first, skip=skip)
        """)
    def get_latest_labels(self, project_id: str, skip: int, first: int):
        return self.labels(project_id=project_id, first=first, skip=skip)

    def labels(self,
               asset_id: str = None,
               asset_status_in: List[str] = None,
               asset_external_id_in: List[str] = None,
               author_in: List[str] = None,
               created_at: str = None,
               created_at_gte: str = None,
               created_at_lte: str = None,
               first: int = None,
               honeypot_mark_gte: float = None,
               honeypot_mark_lte: float = None,
               label_id: str = None,
               project_id: str = None,
               skip: int = 0,
               skipped: bool = None,
               type_in: List[str] = None,
               user_id: str = None):
        """
        Get an array of labels from a project

        Parameters
        ----------
        - asset_id : str
            Identifier of the asset.
        - asset_status_in : list of str, optional (default = None)
            Returned labels should have a status that belongs to that list, if given.
            Possible choices : {'TODO', 'ONGOING', 'LABELED', 'REVIEWED'}
        - asset_external_id_in : list of str, optional (default = None)
            Returned labels should have an external id that belongs to that list, if given.
        - author_in : list of str, optional (default = None)
            Returned labels should have a label whose status belongs to that list, if given.
        - created_at : float, optional (default = None)
            Returned labels should have a label whose creation date is equal to this date.
        - created_at_gt : float, optional (default = None)
            Returned labels should have a label whose creation date is greater than this date.
        - created_at_lt : float, optional (default = None)
            Returned labels should have a label whose creation date is lower than this date.
        - first : int, optional (default = None)
            Maximum number of labels to return.
        - honeypot_mark_gt : float, optional (default = None)
            Returned labels should have a label whose honeypot is greater than this number.
        - honeypot_mark_lt : float, optional (default = None)
            Returned labels should have a label whose honeypot is lower than this number.
        - label_id : str
            Identifier of the label.
        - project_id : str
            Identifier of the project.
        - skip : int, optional (default = None)
            Number of labels to skip (they are ordered by their date of creation, first to last).
        - skipped : bool, optional (default = None)
            Returned labels should have a label which is skipped
        - type_in : list of str, optional (default = None)
            Returned labels should have a label whose type belongs to that list, if given.
        - user_id : str
            Identifier of the user.


        Returns
        -------
        - a result object which contains the query if it was successful, or an error message else.
        """

        formatted_first = first if first else 100
        variables = {
            'where': {
                'id': label_id,
                'asset': {
                    'id': asset_id,
                    'externalIdIn': asset_external_id_in,
                    'statusIn': asset_status_in,
                },
                'project': {
                    'id': project_id,
                },
                'user': {
                    'id': user_id,
                },
                'typeIn': type_in,
                'authorIn': author_in,
                'honeypotMarkGte': honeypot_mark_gte,
                'honeypotMarkLte': honeypot_mark_lte,
                'createdAt': created_at,
                'createdAtGte': created_at_gte,
                'createdAtLte': created_at_lte,
                'skipped': skipped,
            },
            'skip': skip,
            'first': formatted_first,
        }
        result = self.auth.client.execute(GQL_LABELS, variables)
        return format_result('data', result)

    def parse_json_response_for_single_classification(self, json_response):
        """
        Get the names of categories from a json_response, for a single class - classification task
        """
        categories = self.parse_json_response_for_multi_classification(
            json_response)
        if len(categories) == 0:
            return []

        return categories[0]

    def parse_json_response_for_multi_classification(self, json_response):
        """
        Get the names of categories from a json_response, for a multi class - classification task
        """
        formatted_json_response = eval(json_response)
        if 'categories' not in formatted_json_response:
            return []
        categories = formatted_json_response['categories']
        return list(map(lambda category: category['name'], categories))

    def parse_json_response(self, json_response, interface_category):
        if interface_category == 'SINGLECLASS_TEXT_CLASSIFICATION':
            return self.parse_json_response_for_single_classification(json_response)
        if interface_category == 'MULTICLASS_TEXT_CLASSIFICATION':
            return self.parse_json_response_for_multi_classification(json_response)

        return json_response

    def export_labels_as_df(self, project_id: str):
        """
        Get the labels of a project as a pandas DataFrame

        Parameters
        ----------
        - project_id : str

        Returns
        -------
        - labels_df : pandas DataFrame containing the labels.
        """
        projects = QueriesProject(self.auth).projects(project_id)
        assert len(projects) == 1, NO_ACCESS_RIGHT
        project = projects[0]
        if 'interfaceCategory' not in project:
            return pd.DataFrame()

        interface_category = project['interfaceCategory']
        assets = QueriesAsset(self.auth).assets(project_id=project_id)
        labels = [dict(label, **dict((f'asset__{key}', asset[key]) for key in asset))
                  for asset in assets for label in asset['labels']]
        labels_df = pd.DataFrame(labels)
        labels_df['y'] = labels_df['jsonResponse'].apply(
            lambda json_response: self.parse_json_response(json_response, interface_category))
        return labels_df
